#include <J3ML/LinearAlgebra/Matrix4x4.h>
#include <J3ML/LinearAlgebra/Vector4.h>

namespace LinearAlgebra {
    const Matrix4x4 Matrix4x4::Zero = Matrix4x4(0);
    const Matrix4x4 Matrix4x4::Identity = Matrix4x4({1,0,0,0}, {0,1,0,0}, {0,0,1,0}, {0,0,0,1});
    const Matrix4x4 Matrix4x4::NaN = Matrix4x4(NAN);

    Matrix4x4::Matrix4x4(const Vector4 &r1, const Vector4 &r2, const Vector4 &r3, const Vector4 &r4) {
        this->elems[0][0] = r1.x;
        this->elems[0][1] = r1.y;
        this->elems[0][2] = r1.z;
        this->elems[0][3] = r1.w;

        this->elems[1][0] = r2.x;
        this->elems[1][1] = r2.y;
        this->elems[1][2] = r2.z;
        this->elems[1][3] = r2.w;

        this->elems[2][0] = r3.x;
        this->elems[2][1] = r3.y;
        this->elems[2][2] = r3.z;
        this->elems[2][3] = r3.w;

        this->elems[3][0] = r4.x;
        this->elems[3][1] = r4.y;
        this->elems[3][2] = r4.z;
        this->elems[3][3] = r4.w;
    }

    Matrix4x4::Matrix4x4(float val) {
        this->elems[0][0] = val;
        this->elems[0][1] = val;
        this->elems[0][2] = val;
        this->elems[0][3] = val;

        this->elems[1][0] = val;
        this->elems[1][1] = val;
        this->elems[1][2] = val;
        this->elems[1][3] = val;

        this->elems[2][0] = val;
        this->elems[2][1] = val;
        this->elems[2][2] = val;
        this->elems[2][3] = val;

        this->elems[3][0] = val;
        this->elems[3][1] = val;
        this->elems[3][2] = val;
        this->elems[3][3] = val;
    }

    Matrix4x4::Matrix4x4(float m00, float m01, float m02, float m03, float m10, float m11, float m12, float m13,
                         float m20, float m21, float m22, float m23, float m30, float m31, float m32, float m33) {
        this->elems[0][0] = m00;
        this->elems[0][1] = m01;
        this->elems[0][2] = m02;
        this->elems[0][3] = m03;

        this->elems[1][0] = m10;
        this->elems[1][1] = m11;
        this->elems[1][2] = m12;
        this->elems[1][3] = m13;

        this->elems[2][0] = m20;
        this->elems[2][1] = m21;
        this->elems[2][2] = m22;
        this->elems[2][3] = m23;

        this->elems[3][0] = m30;
        this->elems[3][1] = m31;
        this->elems[3][2] = m32;
        this->elems[3][3] = m33;

    }

    Matrix4x4::Matrix4x4(const Quaternion &orientation) {

    }

    void Matrix4x4::SetTranslatePart(float translateX, float translateY, float translateZ) {
        elems[0][3] = translateX;
        elems[1][3] = translateY;
        elems[2][3] = translateZ;
    }

    void Matrix4x4::SetTranslatePart(const Vector3 &offset) {
        elems[0][3] = offset.x;
        elems[1][3] = offset.y;
        elems[2][3] = offset.z;
    }

    void Matrix4x4::SetRotatePart(const Quaternion &q) {
        SetMatrixRotatePart(*this, q);
    }

    void Matrix4x4::SetMatrixRotatePart(Matrix4x4 &m, const Quaternion& q)
    {
        // See e.g. http://www.geometrictools.com/Documentation/LinearAlgebraicQuaternions.pdf .
        const float x = q.x;
        const float y = q.y;
        const float z = q.z;
        const float w = q.w;
        elems[0][0] = 1 - 2*(y*y + z*z); elems[0][1] =     2*(x*y - z*w); elems[0][2] =     2*(x*z + y*w);
        elems[1][0] =     2*(x*y + z*w); elems[1][1] = 1 - 2*(x*x + z*z); elems[1][2] =     2*(y*z - x*w);
        elems[2][0] =     2*(x*z - y*w); elems[2][1] =     2*(y*z + x*w); elems[2][2] = 1 - 2*(x*x + y*y);
    }

    void Matrix4x4::SetRow(int row, const Vector3 &rowVector, float m_r3) {
        SetRow(row, rowVector.x, rowVector.y, rowVector.z, m_r3);
    }

    void Matrix4x4::SetRow(int row, const Vector4 &rowVector) {
        SetRow(row, rowVector.x, rowVector.y, rowVector.z, rowVector.w);
    }

    void Matrix4x4::SetRow(int row, float m_r0, float m_r1, float m_r2, float m_r3) {
        elems[row][0] = m_r0;
        elems[row][1] = m_r1;
        elems[row][2] = m_r2;
        elems[row][3] = m_r3;
    }

    Matrix4x4::Matrix4x4(const Quaternion &orientation, const Vector3 &translation) {
        SetRotatePart(orientation);
        SetTranslatePart(translation);
        SetRow(3, 0, 0, 0, 1);
    }

    Matrix4x4 Matrix4x4::D3DOrthoProjLH(float n, float f, float h, float v) {
        float p00 = 2.f / h; float p01 = 0;       float p02 = 0;           float p03 = 0.f;
        float p10 = 0;       float p11 = 2.f / v; float p12 = 0;           float p13 = 0.f;
        float p20 = 0;       float p21 = 0;       float p22 = 1.f / (f-n); float p23 = n / (n-f);
        float p30 = 0;       float p31 = 0;       float p32 = 0.f;         float p33 = 1.f;

        return {p00, p01, p02, p03, p10, p11, p12, p13, p20, p21, p22, p23, p30, p31, p32, p33};
    }

    /** This function generates an orthographic projection matrix that maps from
	the Direct3D view space to the Direct3D normalized viewport space as follows:

	In Direct3D view space, we assume that the camera is positioned at the origin (0,0,0).
	The camera looks directly towards the positive Z axis (0,0,1).
	The -X axis spans to the left of the screen, +X goes to the right.
	-Y goes to the bottom of the screen, +Y goes to the top.

	After the transformation, we're in the Direct3D normalized viewport space as follows:

	(-1,-1,0) is the bottom-left corner of the viewport at the near plane.
	(1,1,0) is the top-right corner of the viewport at the near plane.
	(0,0,0) is the center point at the near plane.
	Coordinates with z=1 are at the far plane.

	Examples:
		(0,0,n) maps to (0,0,0).
		(0,0,f) maps to (0,0,1).
		(-h/2, -v/2, n) maps to (-1, -1, 0).
		(h/2, v/2, f) maps to (1, 1, 1).
	*/
    Matrix4x4 Matrix4x4::D3DOrthoProjRH(float n, float f, float h, float v)
    {
        // D3DOrthoProjLH and D3DOrthoProjRH differ from each other in that the third column is negated.
        // This corresponds to LH = RH * In, where In is a diagonal matrix with elements [1 1 -1 1].
        float p00 = 2.f / h; float p01 = 0;       float p02 = 0;           float p03 = 0.f;
        float p10 = 0;       float p11 = 2.f / v; float p12 = 0;           float p13 = 0.f;
        float p20 = 0;       float p21 = 0;       float p22 = 1.f / (n-f); float p23 = n / (n-f);
        float p30 = 0;       float p31 = 0;       float p32 = 0.f;         float p33 = 1.f;
    }


}